from asyncio import subprocess
import pandas as pd
from typing import Tuple, List
import csv
import pysam
from sklearn.utils import resample
from sklearn.model_selection import train_test_split
import os
import random
import time

import motif_utils

POS_GLIST = ["APOE", "ABCA7", "KLF16", "LILRB2", "SIGLEC11"]
NEG_GLIST = ["CGB2", "URI1", "PGLS", "AURKC", "GRAMD1A"]


def extract_sequence_from_fasta(
    fasta_filename: str, gene_loc: Tuple[int, int], chr: str = "19"
) -> str:
    """
    From the given FASTA file, get the nucleotide sequence from the start to end location
    (interval: [start, end)) as a string.

    Gets from chromosome 19 by default.
    Makes string all uppercase.
    """
    start_pos, end_pos = gene_loc
    ref = pysam.FastaFile(fasta_filename)
    seq = ref.fetch(chr, start_pos, end_pos)
    return seq.upper()


def get_gene_location(gene_name: str, gene_annot_filename: str) -> Tuple[int, int]:
    """Given the name of a gene and a file name for the gene annotation CSV file, return the start
    and end positions (inclusive) of a gene in the chromosome."""
    df = pd.read_csv(gene_annot_filename)
    mask = df["symbol"] == gene_name
    start_pos = int(df[mask]["start"])
    end_pos = int(df[mask]["end"])
    return start_pos, end_pos


def generate_start_end_windows(
    gene_loc: Tuple[int, int], window_size: int, overlap_size: int
) -> List[Tuple[int, int]]:
    """Get a list of tuples with start and end locations of overlapping windows within a certain gene.
    Interval will be used as [start, end)"""
    start_pos, end_pos = gene_loc
    windows = []
    start = start_pos
    end = start_pos + window_size

    while end <= end_pos:
        windows.append((start, end))
        start = end - overlap_size
        end = start + window_size

        if end >= end_pos:
            # have reached the last bit of the gene but it is shorter than window size
            # just take the remaining gene sequence as our last window
            windows.append((start, end_pos))
            break

    return windows


def get_window_at_loc(loc: int, window_size: int) -> Tuple[int, int]:
    """
    Gets a window of the desired size + 1 centered around the specified location.
    
    If window size is even, gets an equal number of base pairs on each side of location.
    If window size is odd, gets an one more base pair on the left side than the right side.
    """
    half = int(window_size / 2)
    end = loc + half
    if window_size % 2 == 0:
        start = loc - half
    else:
        start = loc - half - 1
    return start, end


def generate_tsv_for_window_ADNI(
    window: Tuple[int, int],
    case_control_filepath: str,
    output_tsv_filepath: str,
    fasta_directory: str,
    tokenize: bool,
    loglevel: bool = False,
) -> None:
    """Create a TSV file with the DNA sequence from each window as first column, and label
    of 1 or 0 as second column for all FASTA files listed in case/control file, with labels
    as indicated in this file."""
    with open(output_tsv_filepath, "w+") as tsvfile:
        writer = csv.writer(tsvfile, delimiter="\t")

        with open(case_control_filepath, newline="") as csvfile:
            csv_reader = csv.reader(csvfile, delimiter="\t")
            progress = 0
            for row in csv_reader:
                # get label for patient (0 for control, 1 for case)
                patient_type = row[2]
                label = 0
                if patient_type == "case":
                    label = 1

                # get FASTA file name, read DNA sequence from file, and tokenize it
                file_id = row[3]
                fasta_filename = (
                    fasta_directory + "ADNI_phased_chr19_" + file_id + ".fa"
                )
                try:
                    dna_seq = extract_sequence_from_fasta(fasta_filename, window)
                    # write line to output TSV file
                    if tokenize:
                        kmer = motif_utils.seq2kmer(dna_seq, 6)
                        writer.writerow([kmer, label]) 
                    else:
                        writer.writerow([dna_seq, label])
                except OSError as not_found:
                    print(fasta_filename, "not found; continuing")

                if progress % 50 == 0:
                    if loglevel:
                        print("progress:", progress, "lines done")
                progress += 1


def introduce_artificial_signal(
    input_tsv_filepath: str, output_tsv_filepath: str, signal_fraction: float
) -> None:
    """Generate a new TSV file with an artificial signal, so that the given fraction of the 
    data points with label 1 have a clearly detectable signal, which in this case is the entire nucleotide
    sequence being A's."""
    with open(output_tsv_filepath, "w+") as output_file:
        writer = csv.writer(output_file, delimiter="\t")

        with open(input_tsv_filepath, newline="") as original_file:
            original_reader = csv.reader(original_file, delimiter="\t")
            for row in original_reader:
                if row[1] == "1" and random.random() <= signal_fraction:
                    new_seq = list(row[0])
                    for i in range(len(new_seq)):
                        if new_seq[i] != " ":
                            new_seq[i] = "A"
                    writer.writerow(["".join(new_seq), row[1]])
                else:
                    writer.writerow(row)


def split_tsv_into_test_train(
    input_tsv_filepath: str,
    train_tsv_filepath: str,
    test_tsv_filepath: str,
    train_fraction: float,
) -> None:
    """Given a TSV file, create two new TSV files with the data in a specified train/test split."""
    df = pd.read_csv(input_tsv_filepath, sep="\t")

    # Separate majority and minority classes
    df_majority = df[df.iloc[:, 1] == 1]
    df_minority = df[df.iloc[:, 1] == 0]

    # Downsample majority class
    df_majority_downsampled = resample(
        df_majority,
        replace=False,  # sample without replacement
        n_samples=df_minority.shape[0],
    )  # to match minority class

    # Combine minority class with downsampled majority class
    df_downsampled = pd.concat([df_majority_downsampled, df_minority])
    X = df_downsampled.iloc[:, 0]
    y = df_downsampled.iloc[:, 1]
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, train_size=train_fraction, stratify=y
    )

    train = pd.concat([X_train, y_train], axis=1)
    test = pd.concat([X_test, y_test], axis=1)
    train.to_csv(train_tsv_filepath, sep="\t", index=False)
    test.to_csv(test_tsv_filepath, sep="\t", index=False)


def run_model_with_multiple_windows(
    windows: List[Tuple[int, int]], run_finetune_path: str, data_dir: str
):
    """Generate data files for all the DNA sequence windows in the list, saving in the specified directory, 
    and run DNABERT model using run_finetune.py file on these generated data files for all the windows."""
    for i in range(len(windows)):
        generate_tsv_for_window_ADNI(
            windows[i],
            "/home/user/ADNI_case_control.txt",
            "{}/output.tsv".format(data_dir),
            "/home/user/fasta1/",
            True
        )

        split_tsv_into_test_train(
            "{}/output.tsv".format(data_dir),
            "{}/train.tsv".format(data_dir),
            "{}/dev.tsv".format(data_dir),
            0.9,
        )

        output_dir = "/home/user/DNABERT_model/apoe_window{}_model".format(i)
        os.system("mkdir -p {}".format(output_dir))

        commands = """
        export KMER=6
        export MODEL_PATH=/home/user/DNABERT_model/apoe_window1_model_v2/6-new-12w-0
        export DATA_PATH={}
        export OUTPUT_PATH={}

        python {} \
            --model_type dna \
            --tokenizer_name=dna$KMER \
            --model_name_or_path $MODEL_PATH \
            --task_name dnaprom \
            --do_train \
            --do_eval \
            --data_dir $DATA_PATH \
            --max_seq_length 100 \
            --per_gpu_eval_batch_size=32   \
            --per_gpu_train_batch_size=32   \
            --learning_rate 2e-4 \
            --num_train_epochs 5.0 \
            --output_dir $OUTPUT_PATH \
            --evaluate_during_training \
            --logging_steps 100 \
            --save_steps 4000 \
            --warmup_percent 0.1 \
            --hidden_dropout_prob 0.1 \
            --overwrite_output \
            --weight_decay 0.01 \
            --n_process 8
        """.format(
            data_dir, output_dir, run_finetune_path
        )

        os.system("conda run -n dnabert-run " + commands)


def run_model_with_single_window(
    window: Tuple[int, int],
    run_finetune_path: str,
    artif_sig: bool,
    signal_percent: int=70
) -> None:

    # create directories depending on whether we are introducing artificial signal
    if artif_sig:
        data_dir = "/home/user/DNABERT_model/data/artificial_signal_{}".format(signal_percent)
        output_dir = "/home/user/DNABERT_model/data/artificial_signal_{}/output".format(signal_percent)
    else:
        data_dir = "/home/user/DNABERT_model/data/runs/run{}".format(int(time.time()))
        output_dir = "/home/user/DNABERT_model/data/runs/run{}/output".format(int(time.time()))
    os.system("mkdir -p {}".format(data_dir))

    generate_tsv_for_window_ADNI(
        window,
        "/home/user/ADNI_case_control.txt",
        "{}/output.tsv".format(data_dir),
        "/home/user/fasta1/",
        True
    )

    if artif_sig: 
        signal_fraction = float(signal_percent) / 100
        introduce_artificial_signal("{}/output.tsv".format(data_dir),
                                    "{}/artif_sig_output.tsv".format(data_dir),
                                    signal_fraction)    
        split_tsv_into_test_train(
            "{}/artif_sig_output.tsv".format(data_dir),
            "{}/train.tsv".format(data_dir),
            "{}/dev.tsv".format(data_dir),
            0.9,
        )
    else:   
        split_tsv_into_test_train(
            "{}/output.tsv".format(data_dir),
            "{}/train.tsv".format(data_dir),
            "{}/dev.tsv".format(data_dir),
            0.9,
        )
        
    commands = """
        export KMER=6
        export MODEL_PATH=/home/user/DNABERT_model/apoe_window1_model_v2/6-new-12w-0
        export DATA_PATH={}
        export OUTPUT_PATH={}

        python {} \
            --model_type dna \
            --tokenizer_name=dna$KMER \
            --model_name_or_path $MODEL_PATH \
            --task_name dnaprom \
            --do_train \
            --do_eval \
            --data_dir $DATA_PATH \
            --max_seq_length 100 \
            --per_gpu_eval_batch_size=32   \
            --per_gpu_train_batch_size=32   \
            --learning_rate 2e-4 \
            --num_train_epochs 5.0 \
            --output_dir $OUTPUT_PATH \
            --evaluate_during_training \
            --logging_steps 100 \
            --save_steps 4000 \
            --warmup_percent 0.1 \
            --hidden_dropout_prob 0.1 \
            --overwrite_output \
            --weight_decay 0.01 \
            --n_process 8
        """.format(data_dir, output_dir, run_finetune_path)

    os.system("conda run -n dnabert-run " + commands)
        


def main():
    gene_locs = get_gene_location("APOE", "/home/user/Alzheimers-genetics-NNs/gene_annot.csv")
    # generate_tsv_for_window_ADNI(gene_locs, "/home/user/ADNI_case_control.txt",
    #                 "/home/user/DNABERT_model/data/raw_seqs.tsv",
    #                 "/home/user/fasta1/")

    # windows = generate_start_end_windows(gene_locs, 100, 50)
    # for i in windows:
    #     print(i)
    # window = get_window_at_loc(44908684, 100)
    # print(gene_locs)
    # run_model_with_multiple_windows([gene_locs], "/home/user/DNABERT/examples/run_finetune.py", "/home/user/DNABERT_model/data")
    # run_model_with_multiple_windows(windows, "/home/user/DNABERT/examples/run_finetune.py", "/home/user/DNABERT_model/data")

    # introduce_artificial_signal("/home/user/DNABERT_model/data/output.tsv",
    #                             "/home/user/DNABERT_model/data/artificial_signal_10/output.tsv",
    #                             0.1)
    # windows = []
    for gene in ["EXOC3L2", "SIGLEC11", "TOMM40"]:
        window = get_gene_location(gene, "/home/user/Alzheimers-genetics-NNs/gene_annot.csv")
        print("{} window: {}".format(gene, window))
        print(window[1]-window[0]+1)
        # windows.append(window)

        # run_model_with_single_window(window,
        #         "/home/user/DNABERT/examples/run_finetune.py",
        #         False)

    # run_model_with_multiple_windows(windows, "/home/user/DNABERT/examples/run_finetune.py", "/home/user/DNABERT_model/data")

    # run_model_with_single_window(gene_locs,
    #         "/home/user/DNABERT_model/data/artificial_signal2/data",
    #         "/home/user/DNABERT_model/data/artificial_signal2/output",
    #         "/home/user/DNABERT/examples/run_finetune.py",
    #         True,
    #         90)

    # window = get_gene_location("EXOC3L2", "/home/user/Alzheimers-genetics-NNs/gene_annot.csv")
    # run_model_with_single_window(window,
    #             "/home/user/DNABERT/examples/run_finetune.py",
    #             False)

    # windows = generate_start_end_windows(gene_locs, 100, 50)
    # for i in windows:
    #     run_model_with_single_window(i,
    #                 "/home/user/DNABERT/examples/run_finetune.py",
    #                 False)



if __name__ == "__main__":
    main()
